#include "TComStat.h"

#if QC_TCOMSTAT

#include <cmath>
#include "TComPic.h"
#include "../../revision.h"

#define LX_TCOMSTAT_LOG_AVER_SEQ			  0x0001
#define LX_TCOMSTAT_LOG_AVER_SLICETYPES	0x0002
#define LX_TCOMSTAT_LOG_AVER_TEMP		    0x0004
#define LX_TCOMSTAT_LOG_FRAME				    0x0008
#define LX_TCOMSTAT_LOG_MAX				      0x000F
#define LX_TCOMSTAT_SLICE_INTER         2 + 1
#define LX_TCOMSTAT_SLICE_ALL	          LX_TCOMSTAT_SLICE_INTER + 1

#ifndef WIN32
#define min(a,b) (a<b? a: b)
#define max(a,b) (a>b? a: b)
#endif

SFrameStat operator + ( const SFrameStat & rSrc1 , const SFrameStat & rSrc2 )
{
	SFrameStat sSum;

	sSum.dAverQP = rSrc1.dAverQP + rSrc2.dAverQP;

	sSum.dPSNRY	= rSrc1.dPSNRY + rSrc2.dPSNRY;
	sSum.dPSNRU = rSrc1.dPSNRU + rSrc2.dPSNRU;
	sSum.dPSNRV = rSrc1.dPSNRV + rSrc2.dPSNRV;

  sSum.dBitsTotal = rSrc1.dBitsTotal + rSrc2.dBitsTotal;
  sSum.dBitsMVField = rSrc1.dBitsMVField + rSrc2.dBitsMVField;
  sSum.dBitsCoeffLuma = rSrc1.dBitsCoeffLuma + rSrc2.dBitsCoeffLuma;
  sSum.dBitsCoeffChroma = rSrc1.dBitsCoeffChroma + rSrc2.dBitsCoeffChroma;
  sSum.dBitsGRP = rSrc1.dBitsGRP + rSrc2.dBitsGRP;
  sSum.dBitsMH = rSrc1.dBitsMH + rSrc2.dBitsMH;

	sSum.dRhoY = rSrc1.dRhoY + rSrc2.dRhoY;
	sSum.dRhoU = rSrc1.dRhoU + rSrc2.dRhoU;
	sSum.dRhoV = rSrc1.dRhoV + rSrc2.dRhoV;
	sSum.dRhoY1 = rSrc1.dRhoY1 + rSrc2.dRhoY1;
	sSum.dRhoU1 = rSrc1.dRhoU1 + rSrc2.dRhoU1;
	sSum.dRhoV1 = rSrc1.dRhoV1 + rSrc2.dRhoV1;

	sSum.dPerIntraRegion	= rSrc1.dPerIntraRegion		+ rSrc2.dPerIntraRegion		;
  sSum.dPerIntraBLRegion = rSrc1.dPerIntraBLRegion + rSrc2.dPerIntraBLRegion;
	sSum.dPerSkippedRegion	= rSrc1.dPerSkippedRegion	+ rSrc2.dPerSkippedRegion	;
	sSum.dPerMergeRegion	= rSrc1.dPerMergeRegion		+ rSrc2.dPerMergeRegion	;
  sSum.dPerInterRegion = rSrc1.dPerInterRegion + rSrc2.dPerInterRegion;

	sSum.dAverDepth			= rSrc1.dAverDepth			+ rSrc2.dAverDepth			;
	sSum.dAverMV[0]			= rSrc1.dAverMV[0]			+ rSrc2.dAverMV[0]			;
	sSum.dAverMV[1]			= rSrc1.dAverMV[1]			+ rSrc2.dAverMV[1]			;

  for( Int nW = 0 ; nW < SHVC_MAX_GRPW ; nW++ )
  {
    sSum.dGRP[nW] = rSrc1.dGRP[nW] + rSrc2.dGRP[nW];
    sSum.dGRPPart[nW][0] = rSrc1.dGRPPart[nW][0] + rSrc2.dGRPPart[nW][0];
    sSum.dGRPPart[nW][1] = rSrc1.dGRPPart[nW][1] + rSrc2.dGRPPart[nW][1];
    for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
    {
      sSum.dGRPDistri_Skip [nW][n] = rSrc1.dGRPDistri_Skip [nW][n] + rSrc2.dGRPDistri_Skip [nW][n];
      sSum.dGRPDistri_Merge[nW][n] = rSrc1.dGRPDistri_Merge[nW][n] + rSrc2.dGRPDistri_Merge[nW][n];
      sSum.dGRPDistri_Inter[nW][n] = rSrc1.dGRPDistri_Inter[nW][n] + rSrc2.dGRPDistri_Inter[nW][n];
    }
  }

  sSum.dMH = rSrc1.dMH + rSrc2.dMH;
  for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
  {
    sSum.dMHDistri[n] = rSrc1.dMHDistri[n] + rSrc2.dMHDistri[n];
  }

  sSum.dUniPred = rSrc1.dUniPred + rSrc2.dUniPred;
  for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
  {
    sSum.dUniPredDistri[n] = rSrc1.dUniPredDistri[n] + rSrc2.dUniPredDistri[n];
  }

	sSum.dLambdaLuma = rSrc1.dLambdaLuma + rSrc2.dLambdaLuma;
  sSum.dLambdaChroma = rSrc1.dLambdaChroma + rSrc2.dLambdaChroma;

	sSum.dTimingFrame	= rSrc1.dTimingFrame	+ rSrc2.dTimingFrame;

	sSum.nFrameNum = rSrc1.nFrameNum + rSrc2.nFrameNum; 

	return( sSum );
}

TComStat::TComStat()
{ 
  m_pStatCurFrame = NULL;
  m_pStatAverTemp0NoI = NULL;
  m_pStatZero = NULL;
  memset( m_pStatPrevFrame , 0 , sizeof( m_pStatPrevFrame ) );
  memset( m_pStatAver , 0 , sizeof( m_pStatAver ) );
  memset( m_pStatAverTemp , 0 , sizeof( m_pStatAverTemp ) );
}

TComStat::~TComStat()
{
  if( m_pStatCurFrame != NULL )
  {
    delete m_pStatCurFrame;
    m_pStatCurFrame = NULL;
  }

  if( m_pStatAverTemp0NoI != NULL )
  {
    delete m_pStatAverTemp0NoI;
    m_pStatAverTemp0NoI = NULL;
  }

  if( m_pStatZero != NULL )
  {
    delete m_pStatZero;
    m_pStatZero = NULL;
  }

  if( m_pStatPrevFrame[0][0] != NULL )
  {
    delete [] m_pStatPrevFrame[0][0];
    memset( m_pStatPrevFrame , 0 , sizeof( m_pStatPrevFrame ) );
  }

  if( m_pStatAver[0] != NULL )
  {
    delete [] m_pStatAver[0];
    memset( m_pStatAver , 0 , sizeof( m_pStatAver ) );
  }

  if( m_pStatAverTemp[0] != NULL )
  {
    delete [] m_pStatAverTemp[0];
    memset( m_pStatAverTemp , 0 , sizeof( m_pStatAverTemp ) );
  }
}

Void TComStat::initStat( TComSPS * pSPS , int nLayerID , int nLogLevel , const char * pSourceFileName , const char * pLogPrefix , TComStat * pBaseStat )
{
  m_pBaseStat = pBaseStat;

  m_pStatCurFrame = new SFrameStat;
  m_pStatAverTemp0NoI = new SFrameStat;
  m_pStatZero = new SFrameStat;

  m_pStatPrevFrame[0][0] = new SFrameStat[3*SHVC_MAX_TLAYERS];
  for( Int type = 0 ; type < 3 ; type++ )
  {
    for( Int n = 0 ; n < SHVC_MAX_TLAYERS ; n++ )
    {
      if( type != 0 || n != 0 )
        m_pStatPrevFrame[type][n] = m_pStatPrevFrame[0][0] + ( type * SHVC_MAX_TLAYERS + n );
    }
  }

  m_pStatAver[0] = new SFrameStat[5];
  for( Int type = 1 ; type < 5 ; type++ )
    m_pStatAver[type] = m_pStatAver[type-1] + 1;

  m_pStatAverTemp[0] = new SFrameStat[SHVC_MAX_TLAYERS];
  for( Int n = 1 ; n < SHVC_MAX_TLAYERS ; n++ )
    m_pStatAverTemp[n] = m_pStatAverTemp[n-1] + 1;

  memset( m_pStatPrevFrame[0][0] , 0 , sizeof( SFrameStat ) * 3 * SHVC_MAX_TLAYERS );
	memset( m_pStatAver[0] , 0 , sizeof( SFrameStat ) * 5 );
	memset( m_pStatAverTemp[0] , 0 , sizeof( SFrameStat ) * SHVC_MAX_TLAYERS );
  memset( m_pStatAverTemp0NoI , 0 , sizeof( SFrameStat ) );
  memset( m_pStatZero , 0 , sizeof( SFrameStat ) );
  m_pStatZero->nFrameNum = 1;

  m_pPrevFrameStat = m_pStatZero;
	m_bFirstFrame = true;

  m_nLayerID = nLayerID;
  char str[16];
  sprintf( str , "L%d" , m_nLayerID );
	m_pSPS = pSPS;
	m_nLogLevel = nLogLevel;
	m_strSourceFileName = pSourceFileName;
	m_strLogPrefix = pLogPrefix;
  m_strLogPrefix += str;
}

Void TComStat::startFrame()
{
	m_TimingFrame.StartTiming();

	// reset frame stat
	memset( m_pStatCurFrame , 0 , sizeof( SFrameStat ) );
}

Void TComStat::endFrame( TComPic * pPic )
{
	if( m_nLogLevel == 0 )
		return;

  // stop timing
  m_pStatCurFrame->dTimingFrame = m_TimingFrame.StopTiming().dUserTime;

  if( pPic != NULL )
  {
    TComSlice & rSlice = *pPic->getSlice( 0 );

    if( m_bFirstFrame )
      m_nFrameSize = m_pSPS->getPicWidthInLumaSamples() * m_pSPS->getPicHeightInLumaSamples();

    // frame stat
    xxStatFrame( *pPic );

    // output log frame
    if( m_nLogLevel & LX_TCOMSTAT_LOG_FRAME )
    {
      char strFileFrame[256];
      sprintf( strFileFrame , "%s_frame.log" , m_strLogPrefix.c_str() );
      FILE * fp = fopen( strFileFrame , "a" );
      if( fp != NULL )
      {
        if( m_bFirstFrame )
          xxOutputStatTitle( fp );
        xxOutputFrameStat( fp , *m_pStatCurFrame , m_pBaseStat == NULL ? *m_pStatZero : *m_pBaseStat->m_pPrevFrameStat , &rSlice );
        fclose( fp );
      }
    }

    // save the stat
    xxSaveCurFrameStat( rSlice.getSliceType() , rSlice.getDepth() );

    m_bFirstFrame = false;
  }
}

Void TComStat::statCU( TComDataCU & rCU )
{
	if( m_nLogLevel == 0 )
		return;

  m_pStatCurFrame->dBitsMVField += rCU.getBitsMVField();
  m_pStatCurFrame->dBitsCoeffLuma += rCU.getBitsCoeffLuma();
  m_pStatCurFrame->dBitsCoeffChroma += rCU.getBitsCoeffChroma();
  m_pStatCurFrame->dBitsGRP += rCU.getBitsGRPW();
  m_pStatCurFrame->dBitsMH += rCU.getBitsMH();

	xxStatCU( rCU , 0 , 0 );
}

void TComStat::destroy()
{
	xxOutputAverStat();

}

Void TComStat::xxOutputStatTitle( FILE * fp , Bool bFullSeq ) const
{
	fprintf( fp , "\n" );
	// 1
  fprintf( fp , "|Version|%-20.20s|%-32.32s|RevN|L|GBrOfMI|LS|POCFN|FL|Q0|Q1|L|r|" , "Date and time" , "Sequence" );
	// 2
	fprintf( fp , "|BitCurL	BitTota	PSNR_Y	PSNR_U	PSNR_V  EncTime|" );
  // 2B
  fprintf( fp , "EnhR|EnhDY|EnhDU|EnhDV|" );
	// 3
	fprintf( fp , "|bMVF|bitY|bitC|bGRP|biMH||rho-Y|rho-U|rho-V|rhoY1|rhoU1|rhoV1|" );			
	// 4A
	fprintf( fp , "|Intr|IrBL|Skip|Merg|Iter|" );
	// 4B
	fprintf( fp , "aDep|avMV0|avMV1|LambdY|LambdC|" );
	// 5 GRP weighting and MH
  fprintf( fp , "|GRP0|GRP1|GRP2|GRP3|MulH|UniP||" );
  // 5 GRP and MH by CU depths
  for( Int nW = 0 ; nW < SHVC_MAX_GRPW ; nW++ )
    fprintf( fp , "G%dSk|G%dMg|G%dIr|G%d2N|G%dn2|" , nW , nW , nW , nW , nW );
  // MH depth
  for( Int nDepth = 0 ; nDepth < SHVC_MAX_CUDEPTH ; nDepth++ )
    fprintf( fp , "MHD%d|" , nDepth );
  // UniPred depth
  for( Int nDepth = 0 ; nDepth < SHVC_MAX_CUDEPTH ; nDepth++ )
    fprintf( fp , "UPD%d|" , nDepth );

  if( bFullSeq )
  {
    // 6 bit distribution
    fprintf( fp , "|bIfm|bTL0|bTL1|bTL2|bTL3|prIfm|prTL0|prTL1|prTL2|prTL3|" );
  }


	fprintf( fp , "\n" );
	fprintf( fp , "1---------------------------------------------------------------------------------------------------" );
	fprintf( fp , "2------------------------------------------------------------------------" );
  fprintf( fp , "3--------------------------------------------------------------" );
	fprintf( fp , "4--------------------------------------------------------" );

  fprintf( fp , "5-------------------------------" );
  // GRP and MH by CU depths
  for( Int nW = 0 ; nW < SHVC_MAX_GRPW ; nW++ )
    fprintf( fp , "-------------------------" );
  for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
    fprintf( fp , "----------" );

  if( bFullSeq )
  {
    fprintf( fp , "6-------------------------------------------------------" );
  }

	fprintf( fp , "\n" );

}

Void TComStat::xxOutputFrameStat( FILE * fp , const SFrameStat & rFrameStat , const SFrameStat & rBaseFrameStat , TComSlice * pCurSlice , Bool bFullSeq ) const
{
	// 1
	fprintf( fp , "|%7s|" , NV_VERSION ); 
	fprintf( fp , "%-20.20s|" , m_TimingFrame.GetTimeString() );
	size_t nPos = m_strSourceFileName.length();
	std::string str = m_strSourceFileName.substr( max( 0 , nPos - 32 ) , nPos );
	fprintf( fp , "%-32.32s|" , str.c_str() );
	fprintf( fp , "%-4.4s|%X|" , SVN_REVISION , m_nLogLevel );
  fprintf( fp , "%d%d%d%d%d%d%d|%d%d|" ,
#if SHVC_QC_GRP
    m_pSPS->getUseGRP() ,
    m_pSPS->getGRPWDBilinear() ,
    m_pSPS->getFastGRP() ,
    0 ,
#if SHVC_QC_GRP_GRPFILTER
    m_pSPS->getUseGRPFilter() ,
#else
    0,
#endif
#else
    0, 0, 0,
#endif
    0,
    0,
    0,
    0 
    );

	if( pCurSlice == NULL )
	{
		// log for average
		fprintf( fp , "%-5d|--|" , rFrameStat.nFrameNum );
	}
	else
	{
		// log for a single frame: POC, slice type and temporal level
		const char cSliceType[3] = { 'B' , 'P' , 'I' };
		fprintf( fp , "%-5d|%c%d|" , pCurSlice->getPOC() , cSliceType[pCurSlice->getSliceType()] , pCurSlice->getDepth() );
	}
	fprintf( fp , "%-2.0lf|%-2.0lf|%d|%d|" , rBaseFrameStat.dAverQP / rBaseFrameStat.nFrameNum , rFrameStat.dAverQP / rFrameStat.nFrameNum ,
    m_nLayerID , m_nResolutionRatio );

	// 2
	fprintf( fp , "|%7.5lf	%7.5lf	%6.3lf	%6.3lf	%6.3lf	%7.2lf|" , 
    rFrameStat.dBitsTotal / rFrameStat.nFrameNum / m_nFrameSize ,
    ( rFrameStat.dBitsTotal / rFrameStat.nFrameNum + rBaseFrameStat.dBitsTotal / rBaseFrameStat.nFrameNum ) / m_nFrameSize ,
		rFrameStat.dPSNRY / rFrameStat.nFrameNum , 
    rFrameStat.dPSNRU / rFrameStat.nFrameNum , 
    rFrameStat.dPSNRV / rFrameStat.nFrameNum ,
		rFrameStat.dTimingFrame / rFrameStat.nFrameNum / 1000
    );

  // 2B
  fprintf( fp , "%4.2lf|%5.2lf|%5.2lf|%5.2lf|" , rFrameStat.dBitsTotal / rFrameStat.nFrameNum 
    / ( rBaseFrameStat.dBitsTotal / rBaseFrameStat.nFrameNum + rFrameStat.dBitsTotal / rFrameStat.nFrameNum ) ,
    rFrameStat.dPSNRY / rFrameStat.nFrameNum - rBaseFrameStat.dPSNRY / rBaseFrameStat.nFrameNum ,
    rFrameStat.dPSNRU / rFrameStat.nFrameNum - rBaseFrameStat.dPSNRU / rBaseFrameStat.nFrameNum ,
    rFrameStat.dPSNRV / rFrameStat.nFrameNum - rBaseFrameStat.dPSNRV / rBaseFrameStat.nFrameNum );

	// 3
	fprintf( fp , "|%4.2lf|%4.2lf|%4.2lf|%4.2lf|%4.2lf|" ,
    // the percentage instead of actual bits/pixel
    rFrameStat.dBitsMVField / rFrameStat.dBitsTotal ,
    rFrameStat.dBitsCoeffLuma / rFrameStat.dBitsTotal ,
    rFrameStat.dBitsCoeffChroma / rFrameStat.dBitsTotal ,
    rFrameStat.dBitsGRP / rFrameStat.dBitsTotal ,
    rFrameStat.dBitsMH / rFrameStat.dBitsTotal );
    
  // 3B
  fprintf( fp , "|%5.3lf|%5.3lf|%5.3lf|%5.3lf|%5.3lf|%5.3lf|" ,
		rFrameStat.dRhoY / rFrameStat.nFrameNum ,
		rFrameStat.dRhoU / rFrameStat.nFrameNum ,
		rFrameStat.dRhoV / rFrameStat.nFrameNum ,
		rFrameStat.dRhoY1 / rFrameStat.nFrameNum ,
		rFrameStat.dRhoU1 / rFrameStat.nFrameNum ,
		rFrameStat.dRhoV1 / rFrameStat.nFrameNum 
    );

	// 4A
	fprintf( fp , "|%4.2lf|%4.2lf|%4.2lf|%4.2lf|%4.2lf|" , 
		rFrameStat.dPerIntraRegion / rFrameStat.nFrameNum ,
    rFrameStat.dPerIntraBLRegion / rFrameStat.nFrameNum ,
		rFrameStat.dPerSkippedRegion / rFrameStat.nFrameNum ,
		rFrameStat.dPerMergeRegion / rFrameStat.nFrameNum ,
		rFrameStat.dPerInterRegion / rFrameStat.nFrameNum );

  // 4B
	fprintf( fp , "%4.2lf|%5.1lf|%5.1lf|%6.1lf|%6.1lf|" ,
    rFrameStat.dAverDepth / rFrameStat.nFrameNum ,
    rFrameStat.dAverMV[0] / rFrameStat.nFrameNum , 
    rFrameStat.dAverMV[1] / rFrameStat.nFrameNum ,
		rFrameStat.dLambdaLuma / rFrameStat.nFrameNum ,
    rFrameStat.dLambdaChroma / rFrameStat.nFrameNum 
    );
	
	// 5
	fprintf( fp , "|%4.2lf|%4.2lf|%4.2lf|%4.2lf|%4.2lf|%4.2lf||" , 
    rFrameStat.dGRP[0] / rFrameStat.nFrameNum , rFrameStat.dGRP[1] / rFrameStat.nFrameNum , rFrameStat.dGRP[2] / rFrameStat.nFrameNum , rFrameStat.dGRP[3] / rFrameStat.nFrameNum ,
    rFrameStat.dMH / rFrameStat.nFrameNum ,
    rFrameStat.dUniPred / rFrameStat.nFrameNum );
  for( Int nW = 0 ; nW < SHVC_MAX_GRPW ; nW++ )
  {
    Double dGRPSkip = 0 , dGRPMerge = 0 , dGRPInter = 0;
    for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
    {
      dGRPSkip += rFrameStat.dGRPDistri_Skip[nW][n];
      dGRPMerge += rFrameStat.dGRPDistri_Merge[nW][n];
      dGRPInter += rFrameStat.dGRPDistri_Inter[nW][n];
    }
    fprintf( fp , "%4.2lf|%4.2lf|%4.2lf|%4.2lf|%4.2lf|" , dGRPSkip / rFrameStat.nFrameNum , dGRPMerge / rFrameStat.nFrameNum , dGRPInter / rFrameStat.nFrameNum ,
      rFrameStat.dGRPPart[nW][0] / rFrameStat.nFrameNum , rFrameStat.dGRPPart[nW][1] / rFrameStat.nFrameNum );
  }
  for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
  {
    fprintf( fp , "%4.2lf|" , rFrameStat.dMHDistri[n] / rFrameStat.nFrameNum );
  }
  for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
  {
    fprintf( fp , "%4.2lf|" , rFrameStat.dUniPredDistri[n] / rFrameStat.nFrameNum );
  }

  if( bFullSeq )
  {
    fprintf( fp , "|%4.2lf|%4.2lf|%4.2lf|%4.2lf|%4.2lf|" , 
      m_pStatAver[I_SLICE]->dBitsTotal / m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->dBitsTotal ,
      m_pStatAverTemp0NoI->dBitsTotal / m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->dBitsTotal ,
      m_pStatAverTemp[1]->dBitsTotal / m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->dBitsTotal ,
      m_pStatAverTemp[2]->dBitsTotal / m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->dBitsTotal ,
      m_pStatAverTemp[3]->dBitsTotal / m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->dBitsTotal 
      );
    Double dAverPSNRY = m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->dPSNRY / m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->nFrameNum;
    fprintf( fp , "%5.2lf|%5.2lf|%5.2lf|%5.2lf|%5.2lf|" , 
      ( m_pStatAver[I_SLICE]->dPSNRY - dAverPSNRY * m_pStatAver[I_SLICE]->nFrameNum ) /  m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->nFrameNum ,
      ( m_pStatAverTemp0NoI->dPSNRY - dAverPSNRY * m_pStatAverTemp0NoI->nFrameNum ) /  m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->nFrameNum ,
      ( m_pStatAverTemp[1]->dPSNRY - dAverPSNRY * m_pStatAverTemp[1]->nFrameNum ) /  m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->nFrameNum ,
      ( m_pStatAverTemp[2]->dPSNRY - dAverPSNRY * m_pStatAverTemp[2]->nFrameNum ) /  m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->nFrameNum ,
      ( m_pStatAverTemp[3]->dPSNRY - dAverPSNRY * m_pStatAverTemp[3]->nFrameNum ) /  m_pStatAver[LX_TCOMSTAT_SLICE_ALL]->nFrameNum 
      );
  }

	fprintf( fp , "\n" );
}

Void TComStat::xxOutputAverStat()
{
	// stat of inter frames
	*m_pStatAver[LX_TCOMSTAT_SLICE_INTER] = *m_pStatAver[P_SLICE] + *m_pStatAver[B_SLICE];
	*m_pStatAver[LX_TCOMSTAT_SLICE_ALL] = *m_pStatAver[LX_TCOMSTAT_SLICE_INTER] + *m_pStatAver[I_SLICE];

	Int nLogOutputLevel = m_nLogLevel;

	// average of frame types
	const char strFileNameSuffix[5][64] = { "AverB.log" , "AverP.log" , "AverI.log" , "AverInter.log" , "AverSeq.log" };
	Bool bAverOutput[5];
	memset( bAverOutput , 0 , sizeof( bAverOutput ) );
	bAverOutput[LX_TCOMSTAT_SLICE_ALL] = ( nLogOutputLevel & LX_TCOMSTAT_LOG_AVER_SEQ ) > 0;
	bAverOutput[I_SLICE] = bAverOutput[P_SLICE] = bAverOutput[B_SLICE] =
		bAverOutput[LX_TCOMSTAT_SLICE_INTER] = ( nLogOutputLevel & LX_TCOMSTAT_LOG_AVER_SLICETYPES ) > 0;
	int i;
	for( i = 0 ; i <= LX_TCOMSTAT_SLICE_ALL ; i++ )
	{
		if( bAverOutput[i] == true && m_pStatAver[i]->nFrameNum > 0 )
		{
			char strFileName[256];
			sprintf( strFileName , "%s_%s" , m_strLogPrefix.c_str() , strFileNameSuffix[i] );
			FILE * fp = fopen( strFileName , "a" );
			if( fp != NULL )
			{
				fseek( fp , 0 , SEEK_END );
				if( ftell( fp ) == 0 )
					xxOutputStatTitle( fp , i == LX_TCOMSTAT_SLICE_ALL );
        xxOutputFrameStat( fp , *m_pStatAver[i] , m_pBaseStat == NULL ? *m_pStatZero : *m_pBaseStat->m_pStatAver[i] , NULL , i == LX_TCOMSTAT_SLICE_ALL );
				fclose( fp );
			}
		}
	}

  //average of temporal levels
  if( nLogOutputLevel & LX_TCOMSTAT_LOG_AVER_TEMP )
  {
    for( i = 0 ; i < SHVC_MAX_TLAYERS ; i++ )
    {
      if( m_pStatAverTemp[i]->nFrameNum > 0 )
      {
        char strFileName[256];
        sprintf( strFileName, "%s_AverTemp%d.log" , m_strLogPrefix.c_str() , i );
        FILE * fp = fopen( strFileName , "a" );
        if( fp != NULL )
        {
          fseek( fp , 0 , SEEK_END );
          if( ftell( fp ) == 0 )
            xxOutputStatTitle( fp );
          xxOutputFrameStat( fp , *m_pStatAverTemp[i] , m_pBaseStat == NULL ? *m_pStatZero : *m_pBaseStat->m_pStatAverTemp[i] );
          fclose( fp );
        }
      }
    }
    if( m_pStatAverTemp0NoI->nFrameNum > 0 )
    {
      char strFileName[256];
      sprintf( strFileName, "%s_AverTemp0NoI.log" , m_strLogPrefix.c_str() );
      FILE * fp = fopen( strFileName , "a" );
      if( fp != NULL )
      {
        fseek( fp , 0 , SEEK_END );
        if( ftell( fp ) == 0 )
          xxOutputStatTitle( fp );
        xxOutputFrameStat( fp , *m_pStatAverTemp0NoI , m_pBaseStat == NULL ? *m_pStatZero : *m_pBaseStat->m_pStatAverTemp0NoI );
        fclose( fp );
      }
    }
  }
}

Void TComStat::xxSaveCurFrameStat( SliceType eCurSliceType , int nCurTempLevel )
{
	*( m_pStatPrevFrame[eCurSliceType][nCurTempLevel] ) = *m_pStatCurFrame;
	*( m_pStatAver[eCurSliceType] ) = *(m_pStatAver[eCurSliceType]) + *m_pStatCurFrame;
	*( m_pStatAverTemp[nCurTempLevel] ) = *( m_pStatAverTemp[nCurTempLevel] ) + *m_pStatCurFrame;
  if( eCurSliceType != I_SLICE && nCurTempLevel == 0 )
    *m_pStatAverTemp0NoI = *m_pStatAverTemp0NoI + *m_pStatCurFrame;
  m_pPrevFrameStat = m_pStatPrevFrame[eCurSliceType][nCurTempLevel];
}

Void TComStat::xxStatFrame( TComPic & rPic )
{
  m_nResolutionRatio = m_pBaseStat == NULL ? 0 : m_nFrameSize / m_pBaseStat->m_nFrameSize;
	rPic.getPSNR( m_pStatCurFrame->dPSNRY , m_pStatCurFrame->dPSNRU , m_pStatCurFrame->dPSNRV );

	m_pStatCurFrame->dBitsTotal = ( float )rPic.getFrameBits();

	m_pStatCurFrame->dRhoY /= m_nFrameSize;
	m_pStatCurFrame->dRhoU /= ( m_nFrameSize >> 2 );
	m_pStatCurFrame->dRhoV /= ( m_nFrameSize >> 2 );
	m_pStatCurFrame->dRhoY1 /= m_nFrameSize;
	m_pStatCurFrame->dRhoU1 /= ( m_nFrameSize >> 2 );
	m_pStatCurFrame->dRhoV1 /= ( m_nFrameSize >> 2 );

  m_pStatCurFrame->dAverQP /= m_nFrameSize;
  m_pStatCurFrame->dLambdaLuma /= m_nFrameSize;
  m_pStatCurFrame->dLambdaChroma /= m_nFrameSize;

  m_pStatCurFrame->dPerIntraRegion /= m_nFrameSize;
  m_pStatCurFrame->dPerIntraBLRegion /= m_nFrameSize;
	m_pStatCurFrame->dPerSkippedRegion /= m_nFrameSize;
	m_pStatCurFrame->dPerMergeRegion /= m_nFrameSize;
  m_pStatCurFrame->dPerInterRegion /= m_nFrameSize;
	m_pStatCurFrame->dAverDepth /= m_nFrameSize;

	m_pStatCurFrame->dAverMV[0] = sqrt( m_pStatCurFrame->dAverMV[0] / m_nFrameSize );
	m_pStatCurFrame->dAverMV[1] = sqrt( m_pStatCurFrame->dAverMV[1] / m_nFrameSize );

  for( Int nW = 0 ; nW < SHVC_MAX_GRPW ; nW++ )
  {
    for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
    {
      m_pStatCurFrame->dGRPDistri_Skip [nW][n] /= m_nFrameSize;
      m_pStatCurFrame->dGRPDistri_Merge[nW][n] /= m_nFrameSize;
      m_pStatCurFrame->dGRPDistri_Inter[nW][n] /= m_nFrameSize;
      m_pStatCurFrame->dGRP[nW] += m_pStatCurFrame->dGRPDistri_Skip[nW][n] 
        + m_pStatCurFrame->dGRPDistri_Merge[nW][n] + m_pStatCurFrame->dGRPDistri_Inter[nW][n];
    }
    m_pStatCurFrame->dGRPPart[nW][0] /= m_nFrameSize;
    m_pStatCurFrame->dGRPPart[nW][1] /= m_nFrameSize;
  }

  m_pStatCurFrame->dMH /= m_nFrameSize;
  for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
    m_pStatCurFrame->dMHDistri[n] /= m_nFrameSize;

  m_pStatCurFrame->dUniPred /= m_nFrameSize;
  for( Int n = 0 ; n < SHVC_MAX_CUDEPTH ; n++ )
    m_pStatCurFrame->dUniPredDistri[n] /= m_nFrameSize;

	m_pStatCurFrame->nFrameNum = 1;
}

// partly derived from Void TEncCu::xEncodeCU( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
Void TComStat::xxStatCU( TComDataCU & rCU, UInt uiAbsPartIdx, UInt uiDepth )
{
	TComPic* pcPic = rCU.getPic();
	Bool bBoundary = false;
	UInt uiLPelX   = rCU.getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ];
	UInt uiRPelX   = uiLPelX + (g_uiMaxCUWidth>>uiDepth)  - 1;
	UInt uiTPelY   = rCU.getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ];
	UInt uiBPelY   = uiTPelY + (g_uiMaxCUHeight>>uiDepth) - 1;

	bBoundary = !( ( uiRPelX < rCU.getSlice()->getSPS()->getPicWidthInLumaSamples() ) && ( uiBPelY < rCU.getSlice()->getSPS()->getPicHeightInLumaSamples() ) );
	if( ( ( uiDepth < rCU.getDepth( uiAbsPartIdx ) ) && ( uiDepth < (g_uiMaxCUDepth-g_uiAddCUDepth) ) ) || bBoundary )
	{
		UInt uiQNumParts = ( pcPic->getNumPartInCU() >> (uiDepth<<1) )>>2;
		for ( UInt uiPartUnitIdx = 0; uiPartUnitIdx < 4; uiPartUnitIdx++, uiAbsPartIdx+=uiQNumParts )
		{
			uiLPelX   = rCU.getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ];
			uiTPelY   = rCU.getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ];

			if( ( uiLPelX < rCU.getSlice()->getSPS()->getPicWidthInLumaSamples() ) && ( uiTPelY < rCU.getSlice()->getSPS()->getPicHeightInLumaSamples() ) )
				xxStatCU( rCU, uiAbsPartIdx, uiDepth+1 );
		}
		return;
	}

	int nWidth = rCU.getWidth( uiAbsPartIdx );
	int nHeight = rCU.getHeight( uiAbsPartIdx );
	int nPartitionSize = nWidth * nHeight;

	// average depth
	m_pStatCurFrame->dAverDepth += uiDepth * nPartitionSize;
  m_pStatCurFrame->dAverQP += rCU.getQP( uiAbsPartIdx ) * nPartitionSize;
  m_pStatCurFrame->dLambdaLuma += rCU.getSlice()->getLambdaLuma() * nPartitionSize;
  m_pStatCurFrame->dLambdaChroma += rCU.getSlice()->getLambdaChroma() * nPartitionSize;

#if INTRA_BL
  if( rCU.isIntraBL( uiAbsPartIdx ) )
#else
  if( false )
#endif
  {
    // intra BL
    m_pStatCurFrame->dPerIntraBLRegion += nPartitionSize;
  }
	else if( rCU.isIntra( uiAbsPartIdx ) )
	{
		// intra region
		m_pStatCurFrame->dPerIntraRegion += nPartitionSize;
	}
  else
  {
    m_pStatCurFrame->dGRPPart[rCU.getGRPW( uiAbsPartIdx )][rCU.getPartitionSize( uiAbsPartIdx ) != SIZE_2Nx2N] += nPartitionSize;

    // MVs and reference frames
    xxStatCUMVRef( rCU , uiAbsPartIdx , uiDepth , REF_PIC_LIST_0 , true );
    if( rCU.getSlice()->getSliceType() == B_SLICE )
      xxStatCUMVRef( rCU , uiAbsPartIdx , uiDepth , REF_PIC_LIST_1 , false );
  }

	// rho
	UInt uiMaxCuWidth = rCU.getSlice()->getSPS()->getMaxCUWidth();
	UInt uiMaxCuHeight = rCU.getSlice()->getSPS()->getMaxCUHeight();
	UInt uiCoffOffset = uiMaxCuWidth * uiMaxCuHeight * uiAbsPartIdx / rCU.getPic()->getNumPartInCU();
	const TCoeff * pCoeffY = rCU.getCoeffY() + uiCoffOffset;
	int x , y;
	double dRhoY = 0 , dRhoU = 0 , dRhoV = 0; 
	double dRhoY1 = 0 , dRhoU1 = 0 , dRhoV1 = 0;
	for( y = 0 ; y < nHeight ; y++ , pCoeffY += nWidth )
	{
		for( x = 0 ; x < nWidth ; x++ )
		{
			dRhoY += pCoeffY[x] == 0;
			dRhoY1 += abs( pCoeffY[x] ) == 1;
		}
	}
	uiCoffOffset >>=2;
	const TCoeff * pCoeffU = rCU.getCoeffCb() + uiCoffOffset;
	const TCoeff * pCoeffV = rCU.getCoeffCr() + uiCoffOffset;
	int nWidthC = nWidth >> 1;
	int nHeightC = nHeight >> 1;
	for( y = 0 ; y < nHeightC ; y++ , pCoeffU += nWidthC , pCoeffV += nWidthC )
	{
		for( x = 0 ; x < nWidthC ; x++ )
		{
			dRhoU += pCoeffU[x] == 0;
			dRhoV += pCoeffV[x] == 0;
			dRhoU1 += abs( pCoeffU[x] ) == 1;
			dRhoV1 += abs( pCoeffV[x] ) == 1;
		}
	}

	m_pStatCurFrame->dRhoY += dRhoY;
	m_pStatCurFrame->dRhoU += dRhoU;
	m_pStatCurFrame->dRhoV += dRhoV;
	m_pStatCurFrame->dRhoY1 += dRhoY1;
	m_pStatCurFrame->dRhoU1 += dRhoU1;
	m_pStatCurFrame->dRhoV1 += dRhoV1;
}

// partly derived from Void TEncEntropy::encodeMvd( TComDataCU* pcCU, UInt uiAbsPartIdx, RefPicList eRefList, Bool bRD )
Void TComStat::xxStatCUMVRef(  TComDataCU & rCU, UInt uiAbsPartIdx , UInt uiDepth , RefPicList eRefList , Bool bCountModeDistri )
{
	UInt uiPartOffset = ( rCU.getPic()->getNumPartInCU() >> ( rCU.getDepth(uiAbsPartIdx) << 1 ) ) >> 2;
  Int nPartitionSize = rCU.getWidth( uiAbsPartIdx ) * rCU.getHeight( uiAbsPartIdx );

	PartSize ePart = rCU.getPartitionSize( uiAbsPartIdx );
	switch ( ePart )
	{
	case SIZE_2Nx2N:
		{
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 1 , bCountModeDistri );
			break;
		}

	case SIZE_2NxN:
		{
      nPartitionSize >>= 1;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.5 , bCountModeDistri );

			uiAbsPartIdx += uiPartOffset << 1;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.5 , bCountModeDistri );
			break;
		}

	case SIZE_Nx2N:
		{
      nPartitionSize >>= 1;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.5 , bCountModeDistri );

			uiAbsPartIdx += uiPartOffset;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.5 , bCountModeDistri );
			break;
		}

	case SIZE_NxN:
		{
      nPartitionSize >>= 2;
			for ( Int iPartIdx = 0; iPartIdx < 4; iPartIdx++ )
			{
        xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.25 , bCountModeDistri );
				uiAbsPartIdx += uiPartOffset;
			}
			break;
		}

  case SIZE_2NxnU:
    {
      nPartitionSize >>= 2;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.25 , bCountModeDistri );

      uiAbsPartIdx += ( uiPartOffset >> 1 );
      nPartitionSize *= 3;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.75 , bCountModeDistri );
      break;
    }

  case SIZE_2NxnD:
    {
      nPartitionSize >>= 2;
      nPartitionSize *= 3;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.75 , bCountModeDistri );

      uiAbsPartIdx += ( uiPartOffset << 1 ) + ( uiPartOffset >> 1 );
      nPartitionSize /= 3;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.25 , bCountModeDistri );
      break;
    }

  case SIZE_nLx2N:
    {
      nPartitionSize >>= 2;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.25 , bCountModeDistri );

      uiAbsPartIdx += ( uiPartOffset >> 2 );
      nPartitionSize *= 3;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.75 , bCountModeDistri );
      break;
    }

  case SIZE_nRx2N:
    {
      nPartitionSize >>= 2;
      nPartitionSize *= 3;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.75 , bCountModeDistri );

      uiAbsPartIdx += uiPartOffset + ( uiPartOffset >> 2 );
      nPartitionSize /= 3;
      xxStatCUBlkMVRef( rCU , uiAbsPartIdx , uiDepth , ePart , eRefList , 0.25 , bCountModeDistri );
      break;
    }

	default:
		assert( 0 );
		break;
	}
}

Void TComStat::xxStatCUBlkMVRef(  TComDataCU & rCU, UInt uiAbsPartIdx , UInt uiDepth , PartSize ePart , RefPicList eRefList , double dSizeFactor , Bool bCountModeDistri )
{
  double dBlockSize = rCU.getWidth( uiAbsPartIdx ) * rCU.getHeight( uiAbsPartIdx ) * dSizeFactor;

  if( bCountModeDistri )
  {
    if( rCU.isSkipped( uiAbsPartIdx ) )
      m_pStatCurFrame->dPerSkippedRegion += dBlockSize;
    else if( rCU.getMergeFlag( uiAbsPartIdx ) )
      m_pStatCurFrame->dPerMergeRegion += dBlockSize;
    else
      m_pStatCurFrame->dPerInterRegion += dBlockSize;
    if( rCU.getCUMvField( REF_PIC_LIST_0 )->getRefIdx( uiAbsPartIdx ) < 0
      || rCU.getCUMvField( REF_PIC_LIST_1 )->getRefIdx( uiAbsPartIdx ) < 0 )
    {
      m_pStatCurFrame->dUniPred += dBlockSize;
      m_pStatCurFrame->dUniPredDistri[uiDepth] += dBlockSize;
    }
#if SHVC_QC_GRP
    if( rCU.isSkipped( uiAbsPartIdx ) )
      m_pStatCurFrame->dGRPDistri_Skip[rCU.getGRPW(uiAbsPartIdx)][uiDepth] += dBlockSize;
    else if( rCU.getMergeFlag( uiAbsPartIdx ) )
      m_pStatCurFrame->dGRPDistri_Merge[rCU.getGRPW(uiAbsPartIdx)][uiDepth] += dBlockSize;
    else
      m_pStatCurFrame->dGRPDistri_Inter[rCU.getGRPW(uiAbsPartIdx)][uiDepth] += dBlockSize;
#endif
  }

  TComCUMvField* pcCUMvField = rCU.getCUMvField( eRefList );
	Int nRefIdx = pcCUMvField->getRefIdx( uiAbsPartIdx );
	if( nRefIdx < 0 )
		return;

	Int nMVx = pcCUMvField->getMv( uiAbsPartIdx ).getHor();
	Int nMVy = pcCUMvField->getMv( uiAbsPartIdx ).getVer();
	m_pStatCurFrame->dAverMV[eRefList] += dBlockSize * ( nMVx * nMVx + nMVy * nMVy );
}

#endif

