#ifndef __TCOMSTAT__
#define __TCOMSTAT__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <cstdio>
#include <memory>
#include <cassert>

#include "TypeDef.h"

#if QC_TCOMSTAT
#include "TComSlice.h"
#include "TComDataCU.h"
#include "tool_timetools.h"

#define SHVC_MAX_CUDEPTH    5
#define SHVC_MAX_TLAYERS    5
#define SHVC_MAX_GRPW       SHVC_QC_GRP_MAXIDX + 1

typedef struct _SFrameStat
{
	// QP
	double	dAverQP;

	// PSNR
	double	dPSNRY;
	double	dPSNRU;
	double	dPSNRV;

	// bits
  double  dBitsTotal;
  double  dBitsMVField;
  double  dBitsCoeffLuma;
  double  dBitsCoeffChroma;
  double  dBitsGRP;
  double  dBitsMH;

	Double 	dRhoY;				// the percentage of non-zero luma coefficients 
	Double 	dRhoU;
	Double 	dRhoV;
	Double 	dRhoY1;       // percentage of luma coefficients which are quantized to 1/-1
	Double 	dRhoU1;
	Double 	dRhoV1;

	// frame info
	double 	dPerIntraRegion;	  // percentage of intra region in a frame (intraBL is not included)
  double  dPerIntraBLRegion;  // percentage of intra BL
	double 	dPerSkippedRegion;	// percentage of skipped region in a frame
	double 	dPerMergeRegion;
  double  dPerInterRegion;
	double 	dAverDepth;			    // average of partitions per CU
	double 	dAverMV[2];			    // average of MVs, pixel/pixel

  double  dGRP[SHVC_MAX_GRPW];
  double  dGRPDistri_Skip[SHVC_MAX_GRPW][SHVC_MAX_CUDEPTH];
  double  dGRPDistri_Merge[SHVC_MAX_GRPW][SHVC_MAX_CUDEPTH];
  double  dGRPDistri_Inter[SHVC_MAX_GRPW][SHVC_MAX_CUDEPTH];
  double  dGRPPart[SHVC_MAX_GRPW][2]; // 2Nx2N or non-2Nx2N

  double  dMH;
  double  dMHDistri[SHVC_MAX_CUDEPTH];  // MH only for P frames and only for skip and merged PUs

  double  dUniPred;
  double  dUniPredDistri[SHVC_MAX_CUDEPTH];

	double	dLambdaLuma;			      // lambda in use
  double  dLambdaChroma;

	// timing
	double	dTimingFrame;		// average coding time per frame

	// frame number;
	int		  nFrameNum;

}SFrameStat;

SFrameStat operator + ( const SFrameStat & rSrc1 , const SFrameStat & rSrc2 );


class TComStat
{

public:

  TComStat() ;
  virtual ~TComStat();

public:

	Void initStat( TComSPS * pSPS , int nLayerID , int nLogLevel , const char * pSourceFileName , const char * pLogPrefix , TComStat * pBaseStat );
	Void startFrame();
	Void endFrame( TComPic * pPic );
	Void statCU( TComDataCU & rCU );
	void destroy();

  const SFrameStat & getCurFrameStat() const
  {
    return( *m_pStatCurFrame );
  }

	const SFrameStat & getPrevFrameStat( SliceType eCurSliceType , int nCurTempLevel = 0 ) const
	{
		return( *m_pStatPrevFrame[eCurSliceType][nCurTempLevel] );
	}

	const char * getLogPrefix()
	{
		return( m_strLogPrefix.c_str() );
	}

  const char * getSourceName() 
  { 
    return( m_strSourceFileName.c_str() );
  }

  Int getFrameSize()
  {
    return( m_nFrameSize );
  }

private:

  TComStat * m_pBaseStat;

  Int m_nLayerID;
  Int m_nResolutionRatio;
	TComSPS * m_pSPS;
	int m_nLogLevel;
	std::string m_strSourceFileName;
	std::string m_strLogPrefix;
  Int m_nFrameSize;

	Bool m_bFirstFrame;  
	SFrameStat * m_pStatCurFrame;
	SFrameStat * m_pStatPrevFrame[3][SHVC_MAX_TLAYERS];	    // slice, temporal level
	SFrameStat * m_pStatAver[5];			                      // B, P, I, Inter, All
	SFrameStat * m_pStatAverTemp[SHVC_MAX_TLAYERS];		      // average of each temporal levels
  SFrameStat * m_pStatAverTemp0NoI;		    
  SFrameStat * m_pStatZero;
  SFrameStat * m_pPrevFrameStat;

	Utilities::CTimeTools m_TimingFrame;

	Void xxStatFrame( TComPic & rPic );
	Void xxOutputStatTitle( FILE * fp , Bool bFullSeq = false ) const;
	Void xxOutputFrameStat( FILE * fp , const SFrameStat & rFrameStat , const SFrameStat & rBaseFrameStat , TComSlice * pCurSlice = NULL , Bool bFullSeq = false ) const;
	Void xxOutputAverStat();
	Void xxSaveCurFrameStat( SliceType eCurSliceType , int nCurTempLevel );
	Void xxStatCU( TComDataCU & rCU , UInt uiAbsPartIdx , UInt uiDepth );
	Void xxStatCUMVRef(  TComDataCU & rCU, UInt uiAbsPartIdx , UInt uiDepth , RefPicList eRefList , Bool bCountModeDistri );
	Void xxStatCUBlkMVRef(  TComDataCU & rCU, UInt uiAbsPartIdx , UInt uiDepth , PartSize ePart , RefPicList eRefList , double dSizeFactor , Bool bCountModeDistri );
};
#endif

#endif // !defined(AFX_TCOMSTAT_H__C79BCAA2_6AC8_4175_A0FE_CF02F5829233__INCLUDED_)

